package sensor.test.app.reconnectingapp;

import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothManager;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.os.IBinder;
import android.util.Log;

import com.mbientlab.metawear.android.BtleService;
import com.mbientlab.metawear.MetaWearBoard;

import bolts.Task;

import static android.content.Context.BLUETOOTH_SERVICE;

public class MetawearDevice implements ServiceConnection {
    private final Context applicationContext;

    private boolean stopped = false;

    private BtleService.LocalBinder serviceBinder;

    private MetaWearBoard mwBoard;

    private final String macAddress;

    private BluetoothManager btManager;
    private BluetoothDevice btDevice;

    private LightSensor sensor;

    public MetawearDevice(final Context applicationContext,
                          final String macAddress) {
        this.applicationContext = applicationContext;
        this.macAddress = macAddress;
    }

    @Override
    public synchronized void onServiceConnected(ComponentName name, IBinder service) {
        Log.i("SENSORTEST","BtleService connected " + this);
        serviceBinder = (BtleService.LocalBinder) service;
        connect(serviceBinder);
    }

    @Override
    public synchronized void onServiceDisconnected(ComponentName name) {
        Log.i("SENSORTEST", "Bluetooth service disconnected " + this);
    }

    public void bindBluetoothService() {
        this.applicationContext.bindService(new Intent(applicationContext, BtleService.class),
                this, Context.BIND_AUTO_CREATE);
    }

    public void unbindBluetoothService() {
        this.applicationContext
                .unbindService(MetawearDevice.this);
    }

    private synchronized void connect(BtleService.LocalBinder serviceBinder) {
        Log.i("SENSORTEST", "Connecting. Using sensor MAC: " + macAddress +", " + this);

        btManager =
                (BluetoothManager) applicationContext.getSystemService(BLUETOOTH_SERVICE);

        if (btDevice != null) {
            serviceBinder.removeMetaWearBoard(btDevice);
        }
        btDevice = btManager.getAdapter().getRemoteDevice(macAddress);
        mwBoard = serviceBinder.getMetaWearBoard(btDevice);

        Log.i("SENSORTEST", "Trying to connect to metawear board" + mwBoard + ", " + this);
        connect();
    }

    private synchronized Task<Void> reconnect(final MetaWearBoard board) {
        // Remove all routes and resources allocated on the board (observers, data processors, timers, and loggers)
        return board.connectAsync().continueWithTask(task -> {
            if (task.isFaulted() && !stopped) {
                Log.e("SENSORTEST","Unable to connect to Metawear. " + mwBoard + ", " + MetawearDevice.this
                                       + "Trying to reconnect." + board +", " + task.getError());
                for (StackTraceElement se : task.getError().getStackTrace()) {
                    Log.i("SENSORTEST", "-- " + se);
                }
                return reconnect(board);
            }
            if (task.isCancelled()) {
                Log.i("SENSORTEST", "Task is canceled " + ", " + MetawearDevice.this);
                return task;
            }
            if (task.isCompleted()) {
                Log.i("SENSORTEST", "Task is completed" + ", " + MetawearDevice.this);
            }
            Log.i("SENSORTEST", "Reconnect Passed" + ", " + MetawearDevice.this);
            return Task.forResult(null);
        });
    }

    private synchronized Task<Void> connect() {
        if(mwBoard.isConnected()) {
            Log.i("SENSORTEST", "Already connected to the mwBoard." + ", " + MetawearDevice.this);
            return Task.forResult(null);
        }

        Log.i("SENSORTEST", "Connecting to BLE device." + ", " + MetawearDevice.this);
        // Remove all routes and resources allocated on the board (observers, data processors, timers, and loggers)
        return mwBoard.connectAsync()
                .continueWithTask((task) -> {
                    if (task.isCancelled()) {
                        Log.i("SENSORTEST", "Task is cancelled." + ", " + MetawearDevice.this);
                        return task;
                    }

                    if (!task.isFaulted()) {
                        Log.i("SENSORTEST", "Connect task passed. " + mwBoard.isConnected() + ", " + MetawearDevice.this);
                        return task;
                    }

                    Log.i("SENSORTEST", "Connect task fault. " + task.getError() + ", " + MetawearDevice.this);

                    for (StackTraceElement se : task.getError().getStackTrace()) {
                        Log.i("SENSORTEST", "-- " + se + ", " + MetawearDevice.this);
                    }

                    return reconnect(mwBoard);
                }).continueWith(task -> {
                    if (!task.isCancelled() && !stopped) {
                        Log.i("SENSORTEST", "Connected. Setting sensors." + ", " + MetawearDevice.this);
                        mwBoard.onUnexpectedDisconnect(status -> {
                            if (!stopped) {
                                Log.e("SENSORTEST", "!!!!!!!!!!!! Unexpected disconnect. "+ ", " + MetawearDevice.this
                                                       + "Status " + status);
                                connect();
                            }
                        });

                        // clear the board
                        mwBoard.tearDown();

                        if (!startSensors()) {
                            // we were not able to start all required sensor modules
                            // start over
                            if (mwBoard.isConnected()) {
                                mwBoard.disconnectAsync().continueWithTask(dTask -> {
                                    if (dTask.isFaulted()) {
                                        Log.e("SENSORTEST","Disconnect faulted, isConnected == " + mwBoard.isConnected() + " - " + dTask.getError());
                                        return dTask;
                                    }
                                    if (dTask.isCancelled()) {
                                        Log.e("SENSORTEST","Disconnect canceled, isConnected == " + mwBoard.isConnected() + " - " + dTask.getError());
                                    }
                                    if (dTask.isCompleted()) {
                                        Log.e("SENSORTEST","Disconnect completed, isConnected == " + mwBoard.isConnected());
                                    }
                                    Log.e("SENSORTEST","Disconnected. isConnected " + mwBoard.isConnected());
                                    return connect();
                                });
                            }
                        }
                    }
                    return null;
                });
    }

    private synchronized boolean startSensors() {
        sensor = new LightSensor(mwBoard);
        // call sensor.stop() here ? or make state machine which will
        // do reset() on initial connection and after reconnect it will
        // setup everything ?
        sensor.configure();

        if (!sensor.addRoute()) {
            return false;
        }

        sensor.start();
        return true;
    }

    // we call stop in our app when leaving it gracefully
    // we will do also reset() as suggested on forum
    // in this sample app we do not call stop
    private synchronized void stopSensors() {
        Log.i("SENSORTEST", "Stopping sensor");
        sensor.stop();
    }

    public synchronized void destroy() {
        stopped = true;
    }
}
